! This file is part of FD4: Four-Dimensional Distributed Dynamic Data structures
!  
! Copyright 2010 Matthias Lieber, ZIH, TU Dresden, Germany
! matthias.lieber@tu-dresden.de
!  
! FD4 is free software: you can redistribute it and/or modify
! it under the terms of the GNU General Public License as published by
! the Free Software Foundation, either version 3 of the License, or
! (at your option) any later version.
!  
! FD4 is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
! GNU General Public License for more details.
!  
! You should have received a copy of the GNU General Public License
! along with FD4.  If not, see <http://www.gnu.org/licenses/>.

!--------------------------------------------------------------!
!                                                              !
!                    Vis5D Support Routines                    !
!                                                              !
!--------------------------------------------------------------!

#include "fd4.h"

!#define FD4_VIS5D_SINGLE_BLOCKSTRUCTURE

!! Write data of an $fd4_domain to Vis5D file.
!!
!! Only one open-write-close sequence can work at the same time,
!! because this method uses global variables (v5d also has this
!! limitation, so there's no chance to make it better).
!!
!! Vis5D coordinate system:
!!>
!! Coordinates:                     Wind directions:
!!
!! z       y                           A
!!   .------->                       V |   N   
!!   |   N                             |W     E
!!   |W     E                          |   S   
!! x |   S                             .------->
!!   V                               W       U
!!<
!! By translating x and y coordinates, following rules are valid for $fd4_vis5d_mod:
!!>
!! Coordinates:                     Wind directions:
!!
!!    A                                A         
!!  y |   N                          V |   N    
!!    |W     E                         |W     E 
!!    |   S                            |   S    
!!    .------->                        .-------> 
!!  z       x                        W       U   
!!<
!! The following translation is applied:
!!>
!!  x_t =  y
!!  y_t = -x
!!<
!!
!! *TODO*:
!!  - solution for problem with v5df.h -> v5dF.h
!!  - user specified date and time
!!
!! @author Matthias Lieber
module fd4_vis5d_mod

  use fd4_globaldef_mod
  use fd4_vartab_mod
  use fd4_domain_mod
  use fd4_util_mod
  
  implicit none
  include 'v5dF.h'
  private
  
  integer(i_k)  :: abnd(0:3,2), numfields
  real, allocatable :: array4(:,:,:)
  integer :: t=-1
  integer :: a4bnd(3,2)
  
  real(r_k), allocatable :: arrayk(:,:,:,:)
  integer(i_k), allocatable :: index(:), step(:)
  type(fd4_domain), pointer :: d
  logical :: do_zswap
  
  !! optional attributes for $fd4_vis5d_open to specify map projection
  !! and vertical coordinate system
  !!
  !!  - http://vis5d.sourceforge.net/doc/ch03.html#CH03SEC1
  !!  - http://vis5d.sourceforge.net/doc/ch03sec2.html
  type fd4_vis5d_attributes
    integer :: projection               !! type of map projection
    real :: proj_args(7)                !! projection arguments
    integer :: vertical                 !! type of vertical coordinate system
    real, pointer :: vert_args(:)       !! vertical coordinate system arguments (*needs to be allocated*)
  end type
    
  
  ! public methods
  public :: fd4_vis5d_open
  public :: fd4_vis5d_write
  public :: fd4_vis5d_close
  
  public :: fd4_vis5d_attributes


contains


  !! Open Vis5D file.
  !!
  !! You must know which variables and how many time steps you will
  !! write when calling this function!
  !!
  !! $cs will be ignored if $attr is present.
  subroutine fd4_vis5d_open(domain, filename, nsteps, nfields, idx, st, err, cs, attr, zswap)
    type(fd4_domain),  intent(in), target :: domain    !! the domain
    character(*),      intent(in)    :: filename       !! name of the file to write
    integer(i_k),      intent(in)    :: nsteps         !! number of time steps to write
    integer(i_k),      intent(in)    :: nfields        !! number of data fields to write
    integer(i_k),      intent(in)    :: idx(nfields)   !! field indices of the data fields to write
    integer(i_k),      intent(in)    :: st(nfields)    !! time step indices of the data fields to write
    integer(i_k),      intent(out)   :: err            !! error status: 0...ok    
    real(r_k),optional,intent(in)    :: cs(3)          !! grid cell size in x, y, z
    type(fd4_vis5d_attributes), &
       target,optional,intent(in)    :: attr           !! additional attributes for map projection and vertical coordinate system
    logical,optional,  intent(in)    :: zswap          !! swap z coordinate indexes in output?
    
    ! vis5d variables
    character(10)  :: varnames(MAXVARS)
    character(111) :: name
    integer :: v, b, i, res
    integer :: dates(MAXTIMES), times(MAXTIMES)
    integer :: numtimes, numvars, row, col, lev, vlev(MAXVARS)
    integer :: compress
    type(fd4_vis5d_attributes), target :: default_attr
    type(fd4_vis5d_attributes), pointer :: used_attr
        
    err = -1
    
    FD4_L1_ASSERT(nfields>MAXVARS,'nfields > MAXVARS')
    FD4_L1_ASSERT(size(domain%vartab)<MAXVAL(idx),'idx out of bounds')

    numtimes = nsteps
    !numvars  = INT(nfields)
    row = INT(domain%dbnd(1,2) - domain%dbnd(1,1) + 1)    !  row increases from North  to South
    col = INT(domain%dbnd(2,2) - domain%dbnd(2,1) + 1)    !  col increases from West   to East
    lev = INT(domain%dbnd(3,2) - domain%dbnd(3,1) + 1)    !  lev increases from bottom to top   
    vlev(:) = lev

    abnd(1:3,:) = domain%dbnd
    abnd(0,:) = 1
    
    varnames(:) = ''   
    
    !! block structure

    v = 1
    varnames(v) = 'blocks'  ! first var is block structure

    !! block weights
    v = v + 1
    varnames(v) = 'weight'
    
    !! user specified variables
    
    ! get variable names
    do i=1,nfields
      if(domain%vartab(idx(i))%nbins > 1) then
        do b=1,MIN(domain%vartab(idx(i))%nbins,99)     ! max 99 bins
          v = v + 1
          varnames(v) = domain%vartab(idx(i))%name(1:5)
          write(varnames(v)(6:7),'(I2)') b
        end do
        abnd(0,2) = MAX(abnd(0,2), domain%vartab(idx(i))%nbins)
      else
        v = v + 1
        varnames(v) = domain%vartab(idx(i))%name(1:7)
      end if
    end do
    
    numvars=v
    
    dates(:) = 7303       ! 2007-10-30
    times(:) = 0
    do i=1,numtimes
      ! 9:55
      times(i) = 95500 + i - 1  ! add a second
    end do
    
    compress = 1          ! Compression mode: 1 byte per grid point

    name = filename
    
    if(present(attr)) then
      used_attr => attr
    else
      allocate(default_attr%vert_args(2))

      default_attr%projection = 0        ! type of map projection: linear, rectangular, generic units
      default_attr%proj_args(1) = REAL(domain%dbnd(2,2))    ! North boundary of 3-D box
      default_attr%proj_args(2) = REAL(domain%dbnd(1,2))    ! West boundary of 3-D box
      !default_attr%proj_args(1) = REAL(domain%dbnd(2,2))+13    ! North boundary of 3-D box
      !default_attr%proj_args(2) = REAL(domain%dbnd(1,2))-52    ! West boundary of 3-D box
      !default_attr%proj_args(1) = REAL(domain%dbnd(2,2))+12    ! North boundary of 3-D box
      !default_attr%proj_args(2) = REAL(domain%dbnd(1,2))-48    ! West boundary of 3-D box
      !default_attr%proj_args(3) = 0.17    ! Increment between rows
      !default_attr%proj_args(4) = 0.17    ! Increment between cols
      default_attr%proj_args(3) = 1.0    ! Increment between rows
      default_attr%proj_args(4) = 1.0    ! Increment between cols

      default_attr%vertical = 0          ! equally spaced levels in generic units
      default_attr%vert_args(1) = REAL(domain%dbnd(3,1))       ! height of bottom level
      default_attr%vert_args(2) = 1.0       ! spacing between levels

      if(present(cs)) then
        default_attr%proj_args(3) = REAL(cs(2))
        default_attr%proj_args(4) = REAL(cs(1))
        default_attr%vert_args(2) = REAL(cs(3))
      end if
      used_attr => default_attr
    end if
    
    if(domain%mpi%rank==0) then
      res = v5dcreate( name, numtimes, numvars, MAX(2,col), MAX(2,row), vlev,   &
                       varnames, times, dates, compress,          &
                       used_attr%projection, used_attr%proj_args, used_attr%vertical, used_attr%vert_args )
      FD4_L1_ASSERT(res==0,'v5dcreate failed')
    end if
    
    
    !allocate(array4(row,col,lev))
    !allocate(array4(-col:-1,1:MAX(2,row),lev))   ! we will translate the data, as described above
    ! used bounds of array4, if row or col is 1, we allocate 2, because Vis5D requires this
    a4bnd(:,1) = (/ -col,   1, 1   /)
    a4bnd(:,2) = (/   -1, row, lev /)
    allocate(array4(-1*MAX(2,col):-1,1:MAX(2,row),lev))   ! we will translate the data, as described above
    allocate(arrayk(abnd(0,2),row,col,lev))

    t=1

    allocate(index(nfields))
    allocate(step(nfields))
    index = idx
    step = st
    numfields = nfields
    d => domain
    
    if(associated(default_attr%vert_args)) then
      deallocate(default_attr%vert_args)
    end if
    
    do_zswap = .false.
    if(present(zswap)) then
      do_zswap = zswap
    end if

    err = 0
    
  end subroutine fd4_vis5d_open

 
 
  !! Write current data of the domain to Vis5D file.
  !!
  !! The optional $idx and $st parameters will overwrite
  !! the global variables. $numfields must be the same as
  !! $nfields in $fd4_vis5d_open.
  subroutine fd4_vis5d_write(err, idx_opt, st_opt)
    integer(i_k),           intent(out)   :: err                !! error status: 0...ok    
    integer(i_k), optional, intent(in)    :: idx_opt(numfields) !! field indices of the data fields to write
    integer(i_k), optional, intent(in)    :: st_opt(numfields)  !! time step indices of the data fields to write

    integer :: b, v, fld, res
    integer(i_k) :: bbnd(3,2), z, y, x
    
    if(present(idx_opt)) index = idx_opt
    if(present(st_opt))  step = st_opt

    bbnd = 0
    err = -1
    if(t<0) return
    
    if(size(d%vartab)<MAXVAL(index)) then
      FD4_L1_ERROR('idx out of bounds: '//i2s(MAXVAL(index)))
      return
    end if

    if(d%mpi%rank==0) then

    !! write block structure
    v = 1
    arrayk = 0.0
    ! loop over all blocks of the domain (also non allocated blocks)
    do z=1,d%blnum(3)
      do y=1,d%blnum(2)
        do x=1,d%blnum(1)
           bbnd(1,:) = (/ d%d(1)%bl(x) - d%d(1)%bl(1) + 1 , d%d(1)%bl(x+1) - d%d(1)%bl(1) /)
           bbnd(2,:) = (/ d%d(2)%bl(y) - d%d(2)%bl(1) + 1 , d%d(2)%bl(y+1) - d%d(2)%bl(1) /)
           bbnd(3,:) = (/ d%d(3)%bl(z) - d%d(3)%bl(1) + 1 , d%d(3)%bl(z+1) - d%d(3)%bl(1) /)
           arrayk(1,bbnd(1,1):bbnd(1,2),bbnd(2,1):bbnd(2,2),bbnd(3,1):bbnd(3,2)) = d%blkmap(x,y,z,d%blkmap_now)
           !arrayk(1,bbnd(1,1):bbnd(1,2),bbnd(2,1):bbnd(2,2),bbnd(3,1):bbnd(3,2)) = d%balance%blkweights(x,y,z)
        end do
      end do
    end do
    ! convert to default real and do translation as described in module docu
    call arrayk_to_array4        
    res = v5dWrite( t, v, array4 )            
    FD4_L1_ASSERT(res==0,'v5dwrite failed')    
    

    !! write weights
    v = v + 1
    arrayk = 0.0
    ! loop over all blocks of the domain (also non allocated blocks)
    do z=1,d%blnum(3)
      do y=1,d%blnum(2)
        do x=1,d%blnum(1)
           bbnd(1,:) = (/ d%d(1)%bl(x) - d%d(1)%bl(1) + 1 , d%d(1)%bl(x+1) - d%d(1)%bl(1) /)
           bbnd(2,:) = (/ d%d(2)%bl(y) - d%d(2)%bl(1) + 1 , d%d(2)%bl(y+1) - d%d(2)%bl(1) /)
           bbnd(3,:) = (/ d%d(3)%bl(z) - d%d(3)%bl(1) + 1 , d%d(3)%bl(z+1) - d%d(3)%bl(1) /)
           arrayk(1,bbnd(1,1):bbnd(1,2),bbnd(2,1):bbnd(2,2),bbnd(3,1):bbnd(3,2)) = &
                           REAL(d%balance%blkweights(x,y,z), r_k) / REAL( PRODUCT(bbnd(:,2)-bbnd(:,1)+1), r_k)
        end do
      end do
    end do
    ! convert to default real and do translation as described in module docu
    call arrayk_to_array4        
    res = v5dWrite( t, v, array4 )            
    FD4_L1_ASSERT(res==0,'v5dwrite failed')    

    end if
    
    
    !! write data fields
    
    do fld=1,numfields
      ! get field from domain
      !arrayk = d%vartab(index(fld))%vnull
      !TODO: face vars have addidtional cell plane, this is currently ignored. can v5d handle this??
#ifdef FD4_VIS5D_AVG_FACEVAR_TO_CENTER
      if( d%vartab(index(fld))%facevar/=0 ) abnd(d%vartab(index(fld))%facevar,:)=abnd(d%vartab(index(fld))%facevar,:)+1
#endif
      call fd4_util_get_array(d, index(fld), step(fld), abnd, arrayk, err, 0)
#ifdef FD4_VIS5D_AVG_FACEVAR_TO_CENTER
      if( d%vartab(index(fld))%facevar/=0 ) abnd(d%vartab(index(fld))%facevar,:)=abnd(d%vartab(index(fld))%facevar,:)-1
      ! averaging of the face values to the center
      call average_facevar_to_center
#endif
      FD4_L1_ASSERT(err/=0,'fd4_util_get_array failed')
      if(d%mpi%rank==0) then
        ! loop over bins
        do b=1,MIN(d%vartab(index(fld))%nbins,99)     ! max 99 bins
          !print *,index(fld), step(fld),b
          ! convert to default real and do translation as described in module docu
          call arrayk_to_array4        
          !array4 = REAL(arrayk(b,:,:,:))
          ! write
          v = v + 1
          res = v5dWrite( t, v, array4 )            
          FD4_L1_ASSERT(res==0,'v5dwrite failed')
        end do
      end if
    end do
    
    t=t+1
    err = 0

  
  contains
    
    ! convert to default real and do translation as described in module docu
    subroutine arrayk_to_array4
      integer :: i, j, k
      if(do_zswap) then
        do k=a4bnd(3,1),a4bnd(3,2)
          do j=a4bnd(2,1),a4bnd(2,2)
            do i=a4bnd(1,1),a4bnd(1,2)
              array4(i,j,k)=REAL(arrayk(1,j,-i,ubound(array4,3)+1-k))
            end do
          end do
        end do        
      else
        do j=a4bnd(2,1),a4bnd(2,2)
          do i=a4bnd(1,1),a4bnd(1,2)
            array4(i,j,:)=REAL(arrayk(1,j,-i,:))
          end do
        end do        
      end if
      ! extra stuff for 2D runs
      ! duplicate the 2D dimension, Vis5D requires dimensions to be >= 2
      if(a4bnd(2,1)==a4bnd(2,2)) then
        do k=a4bnd(3,1),a4bnd(3,2)
          do i=a4bnd(1,1),a4bnd(1,2)
            array4(i,:,k) = array4(i,a4bnd(2,1),k)
          end do
        end do        
      end if
      if(a4bnd(1,1)==a4bnd(1,2)) then
        do k=a4bnd(3,1),a4bnd(3,2)
          do j=a4bnd(2,1),a4bnd(2,2)
            array4(:,j,k) = array4(a4bnd(1,1),j,k)
          end do
        end do        
      end if      
    end subroutine arrayk_to_array4
    
#ifdef FD4_VIS5D_AVG_FACEVAR_TO_CENTER
    ! averaging of the face values to the center
    subroutine average_facevar_to_center
      integer :: i, j, k
      if( d%vartab(index(fld))%facevar==1 ) then
        do i=lbound(arrayk,2),ubound(arrayk,2)-1
          arrayk(:,i,:,:) = 0.5_r_k * ( arrayk(:,i,:,:) + arrayk(:,i+1,:,:) )
        end do
      else if( d%vartab(index(fld))%facevar==2 ) then
        do j=lbound(arrayk,3),ubound(arrayk,3)-1
          arrayk(:,:,j,:) = 0.5_r_k * ( arrayk(:,:,j,:) + arrayk(:,:,j+1,:) )
        end do
      else if( d%vartab(index(fld))%facevar==3 ) then
        do k=lbound(arrayk,4),ubound(arrayk,4)-1
          arrayk(:,:,:,k) = 0.5_r_k * ( arrayk(:,:,:,k) + arrayk(:,:,:,k+1) )
        end do
      end if
    end subroutine average_facevar_to_center
#endif


  end subroutine fd4_vis5d_write


  !! Close Vis5D file.
  subroutine fd4_vis5d_close(err)
    integer(i_k),      intent(out)   :: err            !! error status: 0...ok    
    integer :: res
    
    err = -1

    if(d%mpi%rank==0) then
      res = v5dclose()
      FD4_L1_ASSERT(res==0,'v5dclose failed')
    end if
    
    deallocate(array4)
    deallocate(arrayk)
    deallocate(index)
    deallocate(step)

    t=-1
    d => null()
    err = 0

  end subroutine fd4_vis5d_close


end module fd4_vis5d_mod
